/*  Termcap - a library for managing a database of terminal capabilities.
    Copyright (C) 1991 Norman H. Azadian
    Reworked in 2001 by Mark Elbrecht.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <termcap.h>

#ifdef __DJGPP__
#include <sys/farptr.h>
#include <go32.h>
#endif

/* Pointer to the termcap entry's info.  */
static char *tcap_buf = NULL;

#ifdef __DJGPP__
#define DEFAULT_TERMCAP "/dev/env/DJDIR/share/termcap"
#else
#define DEFAULT_TERMCAP "/etc/termcap"
#endif

#ifdef __DJGPP__
static const char internal_termcap[] =
":am:bs:co#80:li#25:bl=^G:cr=^M:le=^H:kb=^H:"
":ta=^I:bt=\E[Z:it#8"
":do=\E[B:le=\E[D:nd=\E[C:up=\E[A:nw=^M^J:"
":DO=\E[%dB:LE=\E[%dD:RI=\E[%dC:UP=\E[%dA:"
":cm=\E[%i%d;%dH:ho=\E[H:"
":kb=^H:kd=\E[B:kh=\E[1~:kH=\E[4~:kl=\E[D:kr=\E[C:ku=\E[A:"
":kD=\E[3~:kI=\E[2~:kN=\E[6~:kP=\E[5~:kd=\E[B:"
":cd=\E[J:ce=\E[K:cl=\E[H\E[2J:"
":al=\E[L:dl=\E[M:AL=\E[%dL:DL=\E[%dM:"
":sf=\E[S:sr=\E[T:SF=\E[%dS:SR=\E[%dT:"
":ic=\E[@:IC=\E[%d@:dc=\E[P:DC=\E[%dP:ec=\E[%dX:"
":mb=\E[5m:me=\E[m:mk=\E[8m:mr=\E[7m:md=\E[1m:"
":so=\E[7m:se=\E[m:us=\E[4m:ve=\E[v:vi=\E[1v:vs=\E[2v:"
":AX:Co#8:NC#3:pa#64:AB=\E[4%dm:AF=\E[3%dm:"
":k1=\E[[A:k2=\E[[B:k3=\E[[C:k4=\E[[D:k5=\E[[E:k6=\E[17~:"
":k7=\E[18~:k8=\E[19~:k9=\E[20~:k;=\E[21~:k0=\E[21~:"
;

#define HAVE_INTERNAL_TERMCAP
#endif

struct buffer
{
  char *start;
  size_t alloc_size;
  char *line;
  char *line_ptr;
  char done;
  FILE *f;
};

inline static int
is_dir_sep(char ch)
{
  return (ch == '/') || (ch == '\\');
}

inline static int
is_absolute(char *ch)
{
  return is_dir_sep(ch[0]) || (ch[0] != '.' && ch[1] == ':'
                               && is_dir_sep(ch[2]));
}

/* Returns pointing to the character immediately following the capability id.
   Returns NULL if tgetent() has not yet been called successfully or
   if capability not found.  */
static char *find_capability(char *id)
{
  char *p;
  char found;

  if (tcap_buf == NULL)
    return NULL;

  for (p = tcap_buf; *p; ++p)
  {
    if (p[0] == ':')
    {
      found = (p[1] == id[0]) && (p[2] == id[1]);
      if (found)
      {
        p += 3;
        /* Return not found when '@' follows a capability. */
        if (*p == '@')
          return NULL;
        else
          break;
      }
      if (p[1])
        p += 2;
    }
  }
  if (*p == '\0')
    p = NULL;
  return p;
}


/* Read a line from the termcap file and add it to the input buffer.  */
static char *
read_entry(struct buffer *buf)
{
  int ch;
  char *ptr = buf->line_ptr;
  char *end = buf->start + buf->alloc_size;
  FILE *f = buf->f;

  while (1)
  {
    ch = fgetc(f);
    if (ch == EOF)
    {
      buf->done = 1;
      *ptr = '\0';
      return ptr;
    }
    if (ptr >= end)
    {
      char *xstart;
      size_t ptr_delta;

      ptr_delta = ptr - buf->start;
      /* Increasing the allocation by a constant amount results in O(n**2)
         performance because of the time spent copying memory.  Increasing by
         an exponential amount has O(n) performance because of the decreased
         need to move memory around.  */
      buf->alloc_size *= 2;
      xstart = realloc(buf->start, buf->alloc_size + 5);
      /* Adjust pointers to new location.  */
      end = xstart + (end - buf->start);
      buf->line = xstart + (buf->line - buf->start);
      buf->start = xstart;
      ptr = buf->start + ptr_delta;
    }
    *ptr = (char)ch;
    ++ptr;
    if (ch == '\n')
      break;
  }
  return ptr;
}

/* Copy the entry to its final location.  Don't copy empty or commented out
   entries.  */
static char *
copy_entry (char *out, const char *in)
{
  while (*in != ':')
    ++in;
  *out = *in;
  ++in;
  ++out;

  while (*in)
  {
    /* Skip over whitespace and colons that would create an empty entry.  */
    while (isspace(*in) || (*in == ':' && out[-1] == ':'))
      ++in;
    /* Skip entries that start with a period.  */
    if (*in == ':' && in[1] == '.')
    {
      ++in;
      while (*in != ':')
        ++in;
    }
    if (*in == '\0')
      break;
    *out = *in;
    ++in;
    ++out;
  }
  *out = '\0';
  return out;
}

/* Try to match a name to an entry.  */
static int
compare_entry(const char *name, size_t name_len, const char *entry)
{
  size_t i;

  while (1)
  {
    i = 0;
    while (i < name_len && name[i] == *entry)
    {
      ++i;
      ++entry;
    }
    if (i == name_len && (*entry == '|' || *entry == ':'))
      return 0;
    while (!(*entry == '|' || *entry == ':' || *entry == '\0'))
      ++entry;
    if (*entry == ':' || *entry == '\0')
      break;
    ++entry;
  }
  return 1;
}

/* Find an entry and copy its capabilities to OUTBUF.  Reads entries into
   TBUF if neccessary.  */
static int
get_entry(const char *name, struct buffer *tbuf, char *outbuf)
{
  size_t *offset;
  char *entry;
  char *line_end;
  size_t name_len;
  char found;

  found = 0;
  name_len = strlen(name);
  offset = (size_t *)tbuf->start;

  /* Search through the entries already added to the buffer.  */
  while (*offset)
  {
    entry = (char *)(offset + 1);
    if (compare_entry(name, name_len, entry) == 0)
    {
      copy_entry(outbuf, entry);
      return 1;
    }
    offset = (size_t *)(tbuf->start + *offset);
  }

  /* Add entries to the buffer until the entry is found or until all
     entries have been searched.  */
  while (tbuf->done == 0)
  {
    /* Read in a line.  */
    line_end = read_entry(tbuf);
    /* Discard a carriage return preceding the newline.  */
    if (line_end[-2] == '\r')
      --line_end;
    /* Handle line continuation.  */
    if ((line_end - tbuf->line) > 2 && line_end[-2] == '\\')
    {
      tbuf->line_ptr = (tbuf->line[0] != '#') ? (line_end - 2)
                                              : (tbuf->line + 1);
      continue;
    }
    else
    {
      /* Handle comments and blank lines.  */
      if (tbuf->line[0] == '#' || isspace(tbuf->line[0]))
        continue;
    }
    line_end[-1] = '\0';
    /* Set the offset to this entry in case the buffer is searched again
       (to handle the :tc capability).  */
    offset = (size_t *)(tbuf->line - sizeof(size_t));
    *offset = (line_end - tbuf->start);
    line_end += sizeof(size_t);
    entry = tbuf->line;
    tbuf->line = line_end;
    tbuf->line_ptr = line_end;
    /* Is this the entry?  */
    if (compare_entry(name, name_len, entry) == 0)
    {
      found = 1;
      break;
    }
  }
  offset = (size_t *)(tbuf->line - sizeof(size_t));
  *offset = 0;

  /* If the entry was found, copy its contents to the output buffer.  */
  if (found)
    copy_entry(outbuf, entry);

  return found;
}

/* Initialize the buffer structure.  */
static int
init_buffer(struct buffer *tbuf, FILE *f)
{
  size_t *offset;

  tbuf->alloc_size = 2048;
  tbuf->start = malloc(tbuf->alloc_size + 1);
  tbuf->line = tbuf->start;
  tbuf->done = 0;

  offset = (size_t *)tbuf->start;
  *offset = 0;
  tbuf->line += sizeof(size_t);
  tbuf->line_ptr = tbuf->line;
  tbuf->f = f;

  return 1;
}

/* Release memory used by the buffer structure.  */
static int
fini_buffer(struct buffer *tbuf)
{
  free(tbuf->start);
  fclose(tbuf->f);
}

/* Extracts the entry for terminal name into the buffer at bp.
   Bp should be a character array of size 1024 and must be retained through
   all subsequent calls to tgetnum(), tgetflag(), and tgetstr().
   Returns -1 if it cannot open the termcap file, 0 if the terminal name
   given does not have an entry, and 1 if all goes well.
   Looks in the environment for a TERMCAP variable.  If found, and the value
   does not begin with a slash, and the terminal type name is the same as
   the environment string TERM, the TERMCAP string is used instead of reading
   the termcap file.  If it does begin with a slash or a drive letter followed
   by a colon and a slash, the string is used as a pathname rather than
   /etc/termcap.  This can speed up entry into programs that call tgetent(),
   as well as to help debug new terminal descriptions or to make one for your
   terminal if you can't write the file /etc/termcap.  */
int tgetent(char *bp, char *name)
{
  char *termcap;		/* pointer to $TERMCAP string */
  const char *filename;		/* name of file with entry */
  FILE *f;
  char found;
  struct buffer tbuf;
  char *entry;
  const char def_termcap[] = DEFAULT_TERMCAP;

  found = 0;
  f = NULL;
  termcap = getenv("TERMCAP");
  filename = def_termcap;

  if ((termcap != NULL) && (*termcap != '\0'))
  {
#ifdef HAVE_INTERNAL_TERMCAP
    if (strcmp(termcap, "djinternal") == 0)
    {
      copy_entry(bp, internal_termcap);
      found = 1;
    }
    else
#endif
    if (is_absolute(termcap))
      filename = termcap;
    else if (strcmp(name, getenv("TERM")) == 0)
    {
      copy_entry(bp, termcap);
      found = 1;  /* success */
    }
  }

  if (found == 0)
  {
    f = fopen(filename, "rb");
    if (f == NULL)
      return -1;

    init_buffer(&tbuf, f);
    found = get_entry(name, &tbuf, bp);
  }

#ifdef HAVE_INTERNAL_TERMCAP
  if (!found)
  {
    copy_entry(bp, internal_termcap);
    found = 1;
  }
#endif

  if (!found)
    return 1;

  /* for our use in future pkg calls */
  tcap_buf = bp;
  entry = bp;

  /* deal with the :tc= capability */
  while (bp = find_capability("tc"))
  {
    char newname[88];

    strncpy(newname, bp + 1, sizeof newname);
    if (strchr(newname, (char)':') != NULL)
      *(strchr(newname, (char)':')) = '\0';
    tcap_buf = bp - 3;
    found = get_entry(newname, &tbuf, tcap_buf);
  }
  tcap_buf = entry;

  /* Close the file now that the entry is complete.  */
  if (f)
    fini_buffer(&tbuf);

  return found;
}


/* Gets the numeric value of capability id, returning -1 if it is not
   given for the terminal.  We allow the form ":xx=N" although only ":xx#N"
   is legal.  We do this because there exist many examples of such in
   4.3bsd /etc/termcap.  */
int tgetnum(char *id)
{
  int retval;
  char *p;

  p = find_capability(id);
  if ((p == NULL) || ((*p != '#') && (*p != '=')))
    return -1; /* not found, or not numeric */

  retval = 0;
  for (++p; *p != ':'; ++p)
    retval = (retval * 10) + (*p - '0');

  return retval;
}


/* Returns 1 if the specified capability is present in the terminal's entry,
   0 if it is not.  This implementation requires that the capability be
   a boolean one.  */
int tgetflag(char *id)
{
  int retval;
  char *p;

  p = find_capability(id);
  retval = (p != NULL && *p == ':');
  return retval;
}


/* Returns the string value of the capability id, places it in the buffer
   at area, and advances the area pointer [past the terminating '\0' char].
   It decodes the abbreviations for this field described in termcap(5),
   except for cursor addressing and padding information.
   Returns NULL if the capability was not found.  */
char *tgetstr(char *id, char **area)
{
  char *retval;
  char *p;

  p = find_capability(id);
  if (p == NULL || *p != '=')
    return NULL;

  retval = *area;
  for (++p; *p != ':'; ++p)
  {
    if (*p == '\\')
    {
      /* Handle escape sequence.  */
      switch (*++p)
      {
        /* Octal constant.  */
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        {
          unsigned sum;
          char *q;

          q = p;
          sum = (p[0] - '0');
          ++p;
          while (*p >= '0' && *p <= '9' && ((p - q) <= 3))
          {
            sum *= 8;
            sum += (p[0] - '0');
            ++p;
          }
          *(*area)++ = (char)(sum & 0377);
          /** will \200 really end up as \000 like it should ? **/
          /** see termcap(5), top of page 6 **/
          break;
        }

        case '^':
          *(*area)++ = '^';
          break;

        case '\\':
          *(*area)++ = '\\';
          break;

        /* Escape */
        case 'E':
          *(*area)++ = '\033';
          break;

        /* Backspace */
        case 'b':
          *(*area)++ = '\b';
          break;

        /* Formfeed */
        case 'f':
          *(*area)++ = '\f';
          break;

        /* Linefeed */
        case 'n':
          *(*area)++ = '\n';
          break;

        /* Carriage return */
        case 'r':
          *(*area)++ = '\r';
          break;

        /* Tab */
        case 't':
          *(*area)++ = '\t';
          break;

        /* Colon */
        case 'C':
        case 'c':
          *(*area)++ = ':';

        /* Unrecognized.  */
        default:
          *(*area)++ = *p;
          break;
      }
    }
    else if (*p == '^')
    {
      /* Control character */
      char ch;

      ++p;
      if (*p == '?')
        ch = 0x7f;
      else
        ch = toupper(*p) - '@';
      *(*area)++ = ch;
    }
    else
      *(*area)++ = *p;		/* normal */
  }

  /* NULL terminate the string.  */
  *(*area)++ = '\0';

  return retval;
}
